// Function.cs
using System;


namespace TtLisp {
  public abstract class Function : Atom {
    // -- instance -------------------------------------------------------
    private List               lambdaList;
    private SymbolicExpression body;

    protected Function() : this( Nil.NIL, Nil.NIL ) {}
    protected Function( List lambdaList, SymbolicExpression body ) {
      this.lambdaList = lambdaList;
      this.body = body;
    }

    protected List               LambdaList { get { return this.lambdaList; } }
    protected SymbolicExpression Body       { get { return this.body; } }

    public abstract SymbolicExpression Call( Env env, List args );

    protected virtual SymbolicExpression Yield( Env env, List args ) {
      if ( this.LambdaList.Size != args.Size ) {
        throw new InvalidArgumentCountException();
      }
      env.PushStack();
      try {
        List ll = this.LambdaList;
        List arg = args;
        while ( !(ll is Nil) ) {
          env.SymbolTable.LocalIntern( (ISymbol)ll.Car, arg.Car );
          ll = (List)ll.Cdr;
          arg = (List)arg.Cdr;
        }
        return env.Eval( this.Body );
      }
      catch ( InvalidCastException ) {
        throw new InvalidLambdaListException();
      }
      finally {
        env.PopStack();
      }
    }

    public override string ToString() {
      return String.Concat( "#<",
                            this.GetType().Name,
                            " lambda-list:",
                            this.LambdaList.ToString(),
                            " body:",
                            this.Body.ToString(),
                            ">" );
    }

  }

  public class Lambda : Function {
    // -- instance -------------------------------------------------------
    protected Lambda() : base() {}
    public Lambda( List lambdaList, SymbolicExpression body ) : base( lambdaList, body ) {}

    public override SymbolicExpression Call( Env env, List args ) {
      return this.Yield( env, this.EvalEachOfList( env, args ) );
    }

    public List EvalEachOfList( Env env, List list ) {
      if ( list is Nil ) { return list; }
      return new List( env.Eval( list.Car ), list.Cdr is List ? EvalEachOfList( env, (List)list.Cdr ) : list.Cdr );
    }
  }

  public abstract class SpecialForm : Function {
    // -- instance -------------------------------------------------------
    protected SpecialForm() : base() {}
    public SpecialForm( List lambdaList, SymbolicExpression body ) : base( lambdaList, body ) {}
  }

  public class Macro : Function {
    // -- instance -------------------------------------------------------
    protected Macro() : base() {}
    public Macro( List lambdaList, SymbolicExpression body ) : base( lambdaList, body ) {}

    public override SymbolicExpression Call( Env env, List args ) {
      return env.Eval( this.Expand( env, args ) );
    }

    public SymbolicExpression Expand( Env env, List args ) {
      return this.Yield( env, args );
    }
  }

}
