// SystemFunctions.cs
using System;


namespace TtLisp.SystemFunctions {

  public class FunctionCar : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      try { return ((List)args[0]).Car; }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionCdr : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      try { return ((List)args[0]).Cdr; }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionQuote : SpecialForm {
    public override SymbolicExpression Call( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      return args[0];
    }
  }

  public class FunctionCons : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      return new List( args[0], args[1] );
    }
  }

  public class FunctionAtom : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      if ( args[0] is Nil ) { return True.T; }
      return args[0] is Atom ? True.T : SymbolicExpression.NIL;
    }
  }

  public class FunctionEq : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      return args[0].LispEq( args[1] ) ? True.T : SymbolicExpression.NIL;
    }
  }

  public class FunctionProgn : SpecialForm {
    public override SymbolicExpression Call( Env env, List args ) {
      List list = args;
      for (;;) {
        SymbolicExpression _tmp = env.Eval( list.Car );
        if ( list.Cdr is Nil || list.Cdr is Atom ) {
          return _tmp;
        }
        list = (List)list.Cdr;
      }
    }
  }

  public class FunctionIf : SpecialForm {
    public override SymbolicExpression Call( Env env, List args ) {
      if ( args.Size != 2 && args.Size != 3 ) {
        throw new InvalidArgumentCountException();
      }

      if ( env.Eval( args.Car ) is Nil ) {
        return env.Eval( args[2] );
      }
      else {
        return env.Eval( args[1] );
      }
    }
  }

  public class FunctionCond : SpecialForm {
    Function progn = new FunctionProgn();
    public override SymbolicExpression Call( Env env, List args ) {
      if ( args.Size < 1  ) {
        throw new InvalidArgumentCountException();
      }
      try {
        List list = args;
        List cur = (List)list.Car;
        for (;;) {
          if ( !(env.Eval( cur.Car ) is Nil) ) {
            return this.progn.Call( env, cur.Cdr is List ? (List)cur.Cdr : Nil.NIL );
          }
          if ( list.Cdr is Nil || list.Cdr is Atom ) {
            return Nil.NIL;
          }
          list = (List)list.Cdr;
          cur = (List)list.Car;
        }
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionLambda : SpecialForm {
    public override SymbolicExpression Call( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      try { return new Lambda( (List)args[0], args[1] ); }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionMacro : SpecialForm {
    public override SymbolicExpression Call( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      try { return new Macro( (List)args[0], args[1] ); }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionFunction : SpecialForm {
    public override SymbolicExpression Call( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      try { return env.FunctionTable[(ISymbol)args[0]]; }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionBindFunction : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      try {
        Symbol funcsym = (Symbol)args[0];
        Function lambda = (Function)args[1];
        env.FunctionTable.Define( funcsym, lambda );
        return funcsym;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionSet : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      try {
        ISymbol sym = (ISymbol)args[0];
        SymbolicExpression val = args[1];
        env.SymbolTable.Intern( sym, val );
        return val;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionApply : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      try {
        ISymbol funcsym = (ISymbol)args[0];
        List _tmp = (List)args[1];
        List arguments = Nil.NIL;
        for ( int i = _tmp.Size - 1; i >= 0; --i ) {
          arguments = new List( List.Make( new Symbol( "quote" ), _tmp[i] ), arguments );
        }
        return env.FunctionTable[funcsym].Call( env, arguments );
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionFuncall : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size < 1 ) {
        throw new InvalidArgumentCountException();
      }
      try {
        ISymbol funcsym = (ISymbol)args[0];
        List _tmp = (List)args.Cdr;
        List arguments = Nil.NIL;
        for ( int i = _tmp.Size - 1; i >= 0; --i ) {
          arguments = new List( List.Make( new Symbol( "quote" ), _tmp[i] ), arguments );
        }
        return env.FunctionTable[funcsym].Call( env, arguments );
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionSetQ : SpecialForm {
    public override SymbolicExpression Call( Env env, List args ) {
      if ( args.Size != 2 ) {
        throw new InvalidArgumentCountException();
      }
      try {
        ISymbol sym = (ISymbol)args[0];
        SymbolicExpression val = env.Eval( args[1] );
        env.SymbolTable.Intern( sym, val );
        return val;
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionEval : Lambda {
    public override SymbolicExpression Call( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      return env.Eval( env.Eval( args[0] ) );
    }
  }

  public class FunctionList : Lambda {
    public override SymbolicExpression Call( Env env, List args ) {
      return this.EvalEachOfList( env, args );
    }
  }

  public class FunctionMacroExpand : Lambda {
    protected override SymbolicExpression Yield( Env env, List args ) {
      if ( args.Size != 1 ) {
        throw new InvalidArgumentCountException();
      }
      try {
        Macro macro = (Macro)env.FunctionTable[(ISymbol)((List)args[0])[0]];
        return macro.Expand( env, (List)((List)args[0]).Cdr );
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
    }
  }

  public class FunctionLetAste : SpecialForm {
    Function progn = new FunctionProgn();
    public override SymbolicExpression Call( Env env, List args ) {
      env.PushStack();
      try {
        List cur = (List)args[0];
        // cur => (X (Y 100) (Z 200))
        while ( !(cur is Nil) ) {
          if ( cur.Car is ISymbol ) {
            env.SymbolTable.LocalIntern( (ISymbol)cur.Car, Nil.NIL );
          }
          else if ( cur.Car is List ) {
            env.SymbolTable.LocalIntern( (ISymbol)((List)cur.Car)[0], env.Eval( ((List)cur.Car)[1] ) );
          }
          else {
            throw new InvalidArgumentTypeException();
          }
          cur = cur.Cdr is List ? (List)cur.Cdr : Nil.NIL;
        }
        return this.progn.Call( env, args.Cdr is List ? (List)args.Cdr : Nil.NIL );
      }
      catch ( InvalidCastException ) {
        throw new InvalidArgumentTypeException();
      }
      finally {
        env.PopStack();
      }
    }
  }
  
}
